<?php

namespace Drupal\drupalmoduleupgrader;

use Drupal\Core\Plugin\PluginBase;
use Drupal\drupalmoduleupgrader\Utility\Filter\ContainsLogicFilter;
use Pharborist\ClassNode;
use Pharborist\DocCommentNode;
use Pharborist\Functions\FunctionDeclarationNode;
use Pharborist\Parser;
use Pharborist\Filter;
use Pharborist\WhitespaceNode;
use Symfony\Component\Filesystem\Filesystem;
use Symfony\Component\Yaml\Yaml;

/**
 * Base class for converters.
 */
abstract class ConverterBase extends PluginBase implements ConverterInterface {

  /**
   * @return mixed
   *
   * @throws \LogicException if the target module doesn't implement the
   * specified hook, or if the implementation contains logic.
   */
  protected function executeHook(TargetInterface $target, $hook) {
    $indexer = $target->getIndexer('hook');
    if ($indexer->has($hook)) {
      $function = $indexer->get($hook)->get(0);
      if ($function->is(new ContainsLogicFilter)) {
        throw new \LogicException('{target}_{hook} cannot be executed because it contains logic.');
      }
      else {
        $function_name = $function->getName()->getText();
        if (! function_exists($function_name)) {
          eval($function->getText());
        }
        return call_user_func($function_name);
      }
    }
    else {
      throw new \LogicException('{target} does not implement hook_{hook}.');
    }
  }

  /**
   * Creates an empty implementation of a hook.
   *
   * @param TargetInterface $target
   *  The target module.
   * @param string $hook
   *  The hook to implement, without the hook_ prefix.
   *
   * @return \Pharborist\Functions\FunctionDeclarationNode
   *  The hook implementation, appended to the main module file.
   */
  protected function implement(TargetInterface $target, $hook) {
    $function = FunctionDeclarationNode::create($target->id() . '_' . $hook);
    $function->setDocComment(DocCommentNode::create('Implements hook_' . $hook . '().'));

    $module_file = $target->getPath('.module');
    $target->getCodeManager()->open($module_file)->append($function);

    WhitespaceNode::create("\n")->insertBefore($function);
    WhitespaceNode::create("\n")->insertAfter($function);

    return $function;
  }

  /**
   * Writes a file to the target module's directory.
   *
   * @param TargetInterface $target
   *  The target module.
   * @param string $path
   *  The path of the file to write, relative to the module root.
   * @param string $data
   *  The file contents.
   *
   * @return string
   *  The path of the file, including the target's base path.
   */
  public function write(TargetInterface $target, $path, $data) {
    static $fs;
    if (empty($fs)) {
      $fs = new Filesystem();
    }

    $destination_path = $target->getPath($path);
    $fs->dumpFile($destination_path, (string) $data);

    return $destination_path;
  }

  /**
   * Writes a class to the target module's PSR-4 root.
   *
   * @param TargetInterface $target
   *  The target module.
   * @param ClassNode $class
   *  The class to write. The path will be determined from the class'
   *  fully qualified name.
   *
   * @return string
   *  The generated path to the class.
   */
  public function writeClass(TargetInterface $target, ClassNode $class) {
    $class_path = ltrim($class->getName()->getAbsolutePath(), '\\');
    $path = str_replace([ 'Drupal\\' . $target->id(), '\\', ], [ 'src', '/' ], $class_path) . '.php';

    return $this->write($target, $path, $class->parents()->get(0));
  }

  /**
   * Writes out arbitrary data in YAML format.
   *
   * @param TargetInterface $target
   *  The target module.
   * @param string $group
   *  The name of the YAML file. It will be prefixed with the module's machine
   *  name and suffixed with .yml. For example, a group value of 'routing'
   *  will write MODULE.routing.yml.
   * @param array $data
   *  The data to write.
   */
  protected function writeInfo(TargetInterface $target, $group, array $data) {
    $destination = $target->getPath('.' . $group . '.yml');
    file_put_contents($destination, Yaml::dump($data, 4, 2));
  }

  protected function writeService(TargetInterface $target, $service_id, array $service_definition) {
    $services = $target->getServices();
    $services->set($service_id, $service_definition);
    $this->writeInfo($target, 'services', [ 'services' => $services->toArray() ]);
  }

  /**
   * Parses a generated class into a syntax tree.
   *
   * @param string|array $class
   *  The class to parse, either as a string of PHP code or a renderable array.
   *
   * @return \Pharborist\ClassNode
   */
  protected function parse($class) {
    if (is_array($class)) {
      $class = drupal_render($class);
    }
    return Parser::parseSnippet($class)->find(Filter::isInstanceOf('Pharborist\ClassNode'))[0];
  }

}
