<?php

namespace Drupal\drupalmoduleupgrader\Plugin\DMU\Converter\Functions;

use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\drupalmoduleupgrader\TargetInterface;
use Pharborist\ClassMethodCallNode;
use Pharborist\Functions\FunctionCallNode;
use Pharborist\LineCommentBlockNode;
use Pharborist\Node;
use Pharborist\ScalarNode;
use Pharborist\StringNode;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Yaml\Yaml;

/**
 * @Converter(
 *  id = "variable_get",
 *  description = @Translation("Replaces variable_get() calls with Configuration API.")
 * )
 */
class VariableGet extends FunctionCallModifier implements ContainerFactoryPluginInterface {

  /**
   * @var \Drupal\Core\StringTranslation\TranslationInterface
   */
  protected $t;

  /**
   * The default configuration values for config/install/MODULE.settings.yml.
   *
   * @var array
   */
  private $config = [];

  /**
   * The configuration schema for config/schema/MODULE.settings.yml.
   *
   * @var array
   */
  private $mapping = [];

  public function __construct(array $configuration, $plugin_id, $plugin_definition, TranslationInterface $translator) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->t = $translator;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('string_translation')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function convert(TargetInterface $target) {
    parent::convert($target);

    $module = $target->id();

    $schema = [
      $module . '.settings' => [
        'type' => 'mapping',
        'label' => 'Settings',
        'mapping' => $this->mapping,
      ],
    ];

    $this->write($target, 'config/install/' . $module . '.settings.yml', Yaml::dump($this->config, 4, 2));
    $this->write($target, 'config/schema/' . $module . '.schema.yml', Yaml::dump($schema, 4, 2));
  }

  /**
   * {@inheritdoc}
   */
  public function rewrite(FunctionCallNode $call, TargetInterface $target) {
    $arguments = $call->getArguments();

    // If there's a second argument, we can try to extract the default value.
    if (isset($arguments[1])) {
      // We can only extract the default value if the key is a constant scalar string.
      if ($arguments[0] instanceof StringNode) {
        $key = $arguments[0]->toValue();
        $value = NULL;

        // Furthermore, we can only extract the default value if it's a constant scalar value.
        // @todo Support arrays, as long as every array element is a constant scalar,
        // at every level of the array.
        if ($arguments[1] instanceof ScalarNode) {
          $value = $arguments[1]->toValue();
        }
        else {
          $fixMe = $this->t->translate("@FIXME: This variable's value could not be added to your module's default configuration, because it is dynamic (NULL was used instead). Please set a valid default for this variable.");
          LineCommentBlockNode::create($fixMe)->insertBefore($call);
        }

        $this->config[$key] = $value;
        $this->mapping[$key]['type'] = getType($value);
      }
      else {
        $fixMe = $this->t->translate("@FIXME: @key must be added to your module's default configuration.", [
          '@key' => $arguments[0]->getText(),
        ]);
        LineCommentBlockNode::create($fixMe)->insertBefore($call);
      }
    }

    return ClassMethodCallNode::create('\Drupal', 'config')
      ->appendArgument(Node::fromValue($target->id() . '.settings'))
      ->appendMethodCall('get')
      ->appendArgument(clone $call->getArgumentList()->getItem(0));
  }

}
