<?php

namespace Drupal\drupalmoduleupgrader\Plugin\DMU\Converter;

use Drupal\Component\Plugin\PluginManagerInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\drupalmoduleupgrader\Converter\Routing\HookMenu;
use Drupal\drupalmoduleupgrader\Converter\Routing\LinkBinding\LinkBinding;
use Drupal\drupalmoduleupgrader\Converter\Routing\LinkBinding\LocalActionLinkBinding;
use Drupal\drupalmoduleupgrader\Converter\Routing\LinkBinding\LocalTaskLinkBinding;
use Drupal\drupalmoduleupgrader\Converter\Routing\LinkBinding\MenuLinkBinding;
use Drupal\drupalmoduleupgrader\Converter\Routing\LinkIndex;
use Drupal\drupalmoduleupgrader\ConverterBase;
use Drupal\drupalmoduleupgrader\TargetInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * @Converter(
 *  id = "links",
 *  description = @Translation("Converts Drupal 7's hook_menu() links to plugin definitions."),
 *  hook = "menu"
 * )
 */
class Links extends ConverterBase implements ContainerFactoryPluginInterface {

  /**
   * @var \Drupal\Component\Plugin\PluginManagerInterface
   */
  protected $routeConverters;

  /**
   * Constructs a Links object.
   *
   * @param array $configuration
   *   Additional configuration for the plugin.
   * @param string $plugin_id
   *   The plugin ID, will be "Links".
   * @param string $plugin_definition
   *   The plugin definition as derived from the annotations.
   * @param \Drupal\Component\Plugin\PluginManagerInterface $route_converters
   *  The plugin manager for route converters, used by HookMenu.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, PluginManagerInterface $route_converters) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->routeConverters = $route_converters;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('plugin.manager.drupalmoduleupgrader.route')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function convert(TargetInterface $target) {
    $hook_menu = new HookMenu($target, $this->routeConverters);

    $links = [
      'menu' => new LinkIndex(),
      'task' => new LinkIndex(),
      'action' => new LinkIndex(),
      'contextual' => new LinkIndex(),
    ];

    foreach ($hook_menu->getSourceRoutes()->getAllLinks() as $path => $source) {
      /** @var LinkBinding $binding */
      $binding = LinkBinding::create($source, $hook_menu->getDestinationRoute($path));

      // Skip if the converter wasn't able to find a destination.
      $destination = $binding->getDestination();
      if (empty($destination)) {
        continue;
      }

      if ($binding instanceof MenuLinkBinding) {
        $links['menu']->addBinding($binding);
      }
      elseif ($binding instanceof LocalTaskLinkBinding) {
        $links['task']->addBinding($binding);
      }
      elseif ($binding instanceof LocalActionLinkBinding) {
        $links['action']->addBinding($binding);
      }
      elseif ($source->isContextualLink()) {
        $links['contextual']->addBinding($binding);
      }
    }

    $links = array_map(function(LinkIndex $index) {
      return $index->build();
    }, $links);

    foreach ($links['contextual'] as $link) {
      $link['group'] = $target->id();
    }

    foreach ($links as $group => $data) {
      if ($data) {
        $this->writeInfo($target, 'links.' . $group, $data);
      }
    }
  }

}
