<?php

namespace Drupal\drupalmoduleupgrader\Utility\ParametricRewriter;

use Pharborist\ArrayLookupNode;
use Pharborist\Filter;
use Pharborist\Functions\CallNode;
use Pharborist\Functions\ParameterNode;
use Pharborist\Node;
use Pharborist\ObjectMethodCallNode;
use Pharborist\ObjectPropertyNode;
use Pharborist\Operator\AssignNode;
use Pharborist\Operator\BooleanNotNode;
use Pharborist\ScalarNode;
use Pharborist\VariableNode;

/**
 * Base class for parametric rewriters that rewrite property accesses.
 */
class PropertyAccessRewriter extends ParametricRewriterBase {

  /**
   * {@inheritdoc}
   */
  public function rewrite(ParameterNode $parameter) {
    parent::rewrite($parameter);

    $reassignments = parent::getWrites($parameter);
    if (sizeof($reassignments) > 0) {
      throw new \LogicException('Cannot parametrically rewrite a function with parameter reassignment.');
    }

    foreach ($this->getReads($parameter) as $read) {
      $getter = $this->asGetter($read);

      if ($getter instanceof CallNode) {
        $empty = $read->closest(Filter::isInstanceOf('\Pharborist\Functions\EmptyNode'));

        // Calls wrapped by empty() are parse errors in PHP 5.4 and older,
        // because PHP is written by Monty Python.
        if ($empty) {
          // The empty() call is negated, e.g. !empty($foo)
          if ($empty->parent() instanceof BooleanNotNode) {
            $empty->parent()->replaceWith($getter);
          }
          else {
            $empty->replaceWith(BooleanNotNode::fromExpression($getter));
          }
        }
      }
    }

    foreach ($this->getWrites($parameter) as $write) {
      $assignment = $write->closest(Filter::isInstanceOf('\Pharborist\Operator\AssignNode'));

      if ($setter = $this->asSetter($write, $assignment)) {
        $assignment->replaceWith($setter);
      }
    }
  }

  protected function asGetter(Node $read) {
    $key = $this->getKey($read);

    if (isset($this->config['properties'][$key]['get'])) {
      return ObjectMethodCallNode::create($this->getVariable($read), $this->config['properties'][$key]['get']);
    }
  }

  protected function asSetter(Node $write, AssignNode $assignment) {
    $key = $this->getKey($write);

    if (isset($this->config['properties'][$key]['set'])) {
      // @todo If the right operand is a variable with a reference symbol,
      // strip the reference (that's call-time pass by reference, and it's no
      // longer in PHP so it'd be a parse error).
      return ObjectMethodCallNode::create($this->getVariable($write), $this->config['properties'][$key]['set'])->appendArgument($assignment->getRightOperand());
    }
  }

  /**
   * {@inheritdoc}
   */
  protected function getReads(ParameterNode $parameter) {
    return $this->getLookups($parameter)->not([ $this, 'isAssigned' ]);
  }

  /**
   * {@inheritdoc}
   */
  protected function getWrites(ParameterNode $parameter) {
    return $this->getLookups($parameter)->filter([ $this, 'isAssigned' ]);
  }

  /**
   * @param \Pharborist\Functions\ParameterNode $parameter
   *
   * @return \Pharborist\NodeCollection
   */
  protected function getLookups(ParameterNode $parameter) {
    return $parameter
      ->getFunction()
      ->find(Filter::isInstanceOf('\Pharborist\ObjectPropertyNode', '\Pharborist\ArrayLookupNode'))
      ->filter($this->isLookupOf($parameter));
  }

  /**
   * @param \Pharborist\Functions\ParameterNode $parameter
   *
   * @return callable
   */
  protected function isLookupOf(ParameterNode $parameter) {
    return function(Node $node) use ($parameter) {
      $var = $node instanceof ArrayLookupNode ? $node->getArray() : $node->getObject();
      return $var instanceof VariableNode && $var->getName() == $parameter->getName();
    };
  }

  /**
   * Tries to get the root variable of a lookup expression.
   *
   * @param \Pharborist\Node $expression
   *  The lookup expression, like $foo['bar']['baz'] or $account->roles[1].
   *
   * @return \Pharborist\Node
   *
   * @todo Ditch this when #161 lands in Pharborist.
   *
   * @throws \DomainException if the variable cannot be determined.
   */
  protected function getVariable(Node $expression) {
    if ($expression instanceof ObjectPropertyNode) {
      return $expression->getObject();
    }
    elseif ($expression instanceof ArrayLookupNode) {
      $array = $expression->getArray();

      // $foo['bar']['baz']
      if ($array instanceof ArrayLookupNode) {
        return $this->getVariable($array);
      }
      // $foo->bar['baz']
      elseif ($array->parent() instanceof ObjectPropertyNode) {
        return $this->getVariable($array->parent());
      }
    }
    throw new \DomainException('Unable to determine root variable of expression ' . $expression);
  }

  /**
   * @return mixed
   */
  protected function getKey(Node $expression) {
    if ($expression instanceof ObjectPropertyNode) {
      return $expression->getProperty()->getText();
    }
    elseif ($expression instanceof ArrayLookupNode) {
      $key = $expression->getKey();

      if ($key instanceof ScalarNode) {
        return $key->toValue();
      }
      else {
        throw new \UnexpectedValueException();
      }
    }
  }

}
